// This is the part of MPFRCPP, the C++ interface for MPFR library.

// Copyright (C) 2006 Alexey Beshenov <bav.272304@gmail.com>

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
// MA 02110-1301 USA.

#ifndef __BASIC_ARITHMETIC_CPP
#define __BASIC_ARITHMETIC_CPP

//------------------------------------------------------------
// MPFRCPP BASIC ARITHMETIC FUNCTIONS
//------------------------------------------------------------

//------------------------------------------------------------
// Absolute value
//------------------------------------------------------------

MPFR abs (const MPFR& x)
    {
    MPFR y(x.prec(), x.round_mode());
    mpfr_abs (y.mpc_x, x.mpc_x, x.mpc_round_mode);
    return y;
    }


//------------------------------------------------------------
// Square root
//------------------------------------------------------------

MPFR sqrt (const MPFR& x)
    {
    MPFR y(x.prec(), x.round_mode());
    mpfr_sqrt (y.mpc_x, x.mpc_x, x.mpc_round_mode);
    return y;
    }


//------------------------------------------------------------
// Cubic root
//------------------------------------------------------------

MPFR cbrt (const MPFR& x)
    {
    MPFR y(x.prec(), x.round_mode());
    mpfr_cbrt (y.mpc_x, x.mpc_x, x.mpc_round_mode);
    return y;
    }


//------------------------------------------------------------
// k-th root
//------------------------------------------------------------

MPFR root (const MPFR& x, const unsigned long int k)
    {
    MPFR y(x.prec(), x.round_mode());
    mpfr_root (y.mpc_x, x.mpc_x, k, x.mpc_round_mode);
    return y;
    }


inline MPFR root (const MPFR& x, const unsigned int k) { return root(x, static_cast<unsigned long int>(k)); }

//------------------------------------------------------------
// Square
//------------------------------------------------------------

MPFR sqr (const MPFR& x)
    {
    MPFR y(x.prec(), x.round_mode());
    mpfr_sqr (y.mpc_x, x.mpc_x, x.mpc_round_mode);
    return y;
    }


//------------------------------------------------------------
// mpfr_neg
//------------------------------------------------------------

MPFR MPFR::operator- () const
    {
    MPFR y(mpfr_get_prec(mpc_x), round_mode());
    mpfr_neg (y.mpc_x, mpc_x, y.mpc_round_mode);
    return y;
    }


//------------------------------------------------------------
// Addition
//------------------------------------------------------------

MPFR operator+ (const MPFR &x, const MPFR &y)
    {
    mp_prec_t prec; std::float_round_style rnd;
    if (x.prec() >= y.prec()) { prec = x.prec(); }
    else { prec = y.prec(); }
    if (x.round_mode() == y.round_mode()) { rnd = x.round_mode(); }
    else { rnd = MPFR::default_round_mode(); }
    MPFR z(prec, rnd);
    mpfr_add (z.mpc_x, x.mpc_x, y.mpc_x, MPFR::frs2mprnd(rnd));
    return z;
    }


MPFR operator+ (const MPFR &x, const unsigned long int &y)
    {
    MPFR z(x.prec(), x.round_mode());
    mpfr_add_ui (z.mpc_x, x.mpc_x, y, x.mpc_round_mode);
    return z;
    }


inline MPFR operator+ (const MPFR &x, const unsigned int &y) { return x+static_cast<unsigned long int>(y); }
inline MPFR operator+ (const unsigned long int &x, const MPFR &y) { return y+x; }
inline MPFR operator+ (const unsigned int &x, const MPFR &y) { return y+static_cast<unsigned long int>(x); }

MPFR operator+ (const MPFR &x, const long int &y)
    {
    MPFR z(x.prec(), x.round_mode());
    mpfr_add_si (z.mpc_x, x.mpc_x, y, x.mpc_round_mode);
    return z;
    }


inline MPFR operator+ (const MPFR &x, const int &y) { return x+static_cast<long int>(y); }
inline MPFR operator+ (const long int &x, const MPFR &y) { return y+x; }
inline MPFR operator+ (const int &x, const MPFR &y) { return y+static_cast<long int>(x); }

MPFR operator+ (const MPFR &x, const mpz_t &y)
    {
    MPFR z(x.prec(), x.round_mode());
    mpfr_add_z (z.mpc_x, x.mpc_x, y, x.mpc_round_mode);
    return z;
    }


inline MPFR operator+ (const mpz_t &x, const MPFR &y) { return y+x; }

MPFR operator+ (const MPFR &x, const mpq_t &y)
    {
    MPFR z(x.prec(), x.round_mode());
    mpfr_add_q (z.mpc_x, x.mpc_x, y, x.mpc_round_mode);
    return z;
    }


inline MPFR operator+ (const mpq_t &x, const MPFR &y) { return y+x; }

//------------------------------------------------------------
// mpfr_sub
//------------------------------------------------------------

MPFR operator- (const MPFR &x, const MPFR &y)
    {
    mp_prec_t prec; std::float_round_style rnd;
    if (x.prec() >= y.prec()) { prec = x.prec(); }
    else { prec = y.prec(); }
    if (x.round_mode() == y.round_mode()) { rnd = x.round_mode(); }
    else { rnd = MPFR::default_round_mode(); }
    MPFR z(prec, rnd);
    mpfr_sub (z.mpc_x, x.mpc_x, y.mpc_x, MPFR::frs2mprnd(rnd));
    return z;
    }


MPFR operator- (const MPFR &x, const unsigned long int &y)
    {
    MPFR z(x.prec(), x.round_mode());
    mpfr_sub_ui (z.mpc_x, x.mpc_x, y, x.mpc_round_mode);
    return z;
    }


inline MPFR operator- (const MPFR &x, const unsigned int &y) { return x-static_cast<unsigned long int>(y); }

MPFR operator- (const unsigned long int &x, const MPFR &y)
    {
    MPFR z(y.prec(), y.round_mode());
    mpfr_ui_sub (z.mpc_x, x, y.mpc_x, y.mpc_round_mode);
    return z;
    }


inline MPFR operator- (const unsigned int &x, const MPFR &y) { return static_cast<unsigned long int>(x)-y; }

MPFR operator- (const MPFR &x, const long int &y)
    {
    MPFR z(x.prec(), x.round_mode());
    mpfr_sub_si (z.mpc_x, x.mpc_x, y, x.mpc_round_mode);
    return z;
    }


inline MPFR operator- (const MPFR &x, const int &y) { return x-static_cast<long int>(y); }

MPFR operator- (const long int &x, const MPFR &y)
    {
    MPFR z(y.prec(), y.round_mode());
    mpfr_si_sub (z.mpc_x, x, y.mpc_x, y.mpc_round_mode);
    return z;
    }


inline MPFR operator- (const int &x, const MPFR &y) { return y-static_cast<long int>(x); }

MPFR operator- (const MPFR &x, const mpz_t &y)
    {
    MPFR z(x.prec(), x.round_mode());
    mpfr_sub_z (z.mpc_x, x.mpc_x, y, x.mpc_round_mode);
    return z;
    }


MPFR operator- (const MPFR &x, const mpq_t &y)
    {
    MPFR z(x.prec(), x.round_mode());
    mpfr_sub_q (z.mpc_x, x.mpc_x, y, x.mpc_round_mode);
    return z;
    }


//------------------------------------------------------------
// Multiplication
//------------------------------------------------------------

MPFR operator* (const MPFR &x, const MPFR &y)
    {
    mp_prec_t prec; std::float_round_style rnd;
    if (x.prec() >= y.prec()) { prec = x.prec(); }
    else { prec = y.prec(); }
    if (x.round_mode() == y.round_mode()) { rnd = x.round_mode(); }
    else { rnd = MPFR::default_round_mode(); }
    MPFR z(prec, rnd);
    mpfr_mul (z.mpc_x, x.mpc_x, y.mpc_x, MPFR::frs2mprnd(rnd));
    return z;
    }


MPFR operator* (const MPFR &x, const unsigned long int &y)
    {
    MPFR z(x.prec(), x.round_mode());
    mpfr_mul_ui (z.mpc_x, x.mpc_x, y, x.mpc_round_mode);
    return z;
    }


inline MPFR operator* (const MPFR &x, const unsigned int &y) { return x*static_cast<unsigned long int>(y); }
inline MPFR operator* (const unsigned long int &x, const MPFR &y) { return y*x; }
inline MPFR operator* (const unsigned int &x, const MPFR &y) { return y*static_cast<unsigned long int>(x); }

MPFR operator* (const MPFR &x, const long int &y)
    {
    MPFR z(x.prec(), x.round_mode());
    mpfr_mul_si (z.mpc_x, x.mpc_x, y, x.mpc_round_mode);
    return z;
    }


inline MPFR operator* (const MPFR &x, const int &y) { return x*static_cast<long int>(y); }
inline MPFR operator* (const long int &x, const MPFR &y) { return y*x; }
inline MPFR operator* (const int &x, const MPFR &y) { return y*static_cast<long int>(x); }

MPFR operator* (const MPFR &x, const mpz_t &y)
    {
    MPFR z(x.prec(), x.round_mode());
    mpfr_mul_z (z.mpc_x, x.mpc_x, y, x.mpc_round_mode);
    return z;
    }


inline MPFR operator* (const mpz_t &y, const MPFR &x) { return y*x; }

MPFR operator* (const MPFR &x, const mpq_t &y)
    {
    MPFR z(x.prec(), x.round_mode());
    mpfr_mul_q (z.mpc_x, x.mpc_x, y, x.mpc_round_mode);
    return z;
    }


inline MPFR operator* (const mpq_t &x, const MPFR &y) { return y*x; }

//------------------------------------------------------------
// Division
//------------------------------------------------------------

MPFR operator/ (const MPFR &x, const MPFR &y)
    {
    mp_prec_t prec; std::float_round_style rnd;
    if (x.prec() >= y.prec()) { prec = x.prec(); }
    else { prec = y.prec(); }
    if (x.round_mode() == y.round_mode()) { rnd = x.round_mode(); }
    else { rnd = MPFR::default_round_mode(); }
    MPFR z(prec, rnd);
    mpfr_div (z.mpc_x, x.mpc_x, y.mpc_x, MPFR::frs2mprnd(rnd));
    return z;
    }


MPFR operator/ (const MPFR &x, const unsigned long int &y)
    {
    MPFR z(x.prec(), x.round_mode());
    mpfr_div_ui (z.mpc_x, x.mpc_x, y, x.mpc_round_mode);
    return z;
    }


inline MPFR operator/ (const MPFR &x, const unsigned int &y) { return x/static_cast<unsigned long int>(y); }

MPFR operator/ (const unsigned long int &x, const MPFR &y)
    {
    MPFR z(y.prec(), y.round_mode());
    mpfr_ui_div (z.mpc_x, x, y.mpc_x, y.mpc_round_mode);
    return z;
    }


inline MPFR operator/ (const unsigned int &y, const MPFR &x) { return static_cast<unsigned long int>(y)/x; }

MPFR operator/ (const MPFR &x, const long int &y)
    {
    MPFR z(x.prec(), x.round_mode());
    mpfr_div_si (z.mpc_x, x.mpc_x, y, x.mpc_round_mode);
    return z;
    }


inline MPFR operator/ (const MPFR &x, const int &y) { return x/static_cast<long int>(y); }

MPFR operator/ (const long int &x, const MPFR &y)
    {
    MPFR z(y.prec(), y.round_mode());
    mpfr_si_div (z.mpc_x, x, y.mpc_x, y.mpc_round_mode);
    return z;
    }


inline MPFR operator/ (const int &y, const MPFR &x) { return static_cast<long int>(y)/x; }

MPFR operator/ (const MPFR &x, const mpz_t &y)
    {
    MPFR z(x.prec(), x.round_mode());
    mpfr_div_z (z.mpc_x, x.mpc_x, y, x.mpc_round_mode);
    return z;
    }


MPFR operator/ (const MPFR &x, const mpq_t &y)
    {
    MPFR z(x.prec(), x.round_mode());
    mpfr_div_q (z.mpc_x, x.mpc_x, y, x.mpc_round_mode);
    return z;
    }


//------------------------------------------------------------
// Power
//------------------------------------------------------------

MPFR pow (const MPFR &x, const MPFR &y)
    {
    mp_prec_t prec; std::float_round_style rnd;
    if (x.prec() >= y.prec()) { prec = x.prec(); }
    else { prec = y.prec(); }
    if (x.round_mode() == y.round_mode()) { rnd = x.round_mode(); }
    else { rnd = MPFR::default_round_mode(); }
    MPFR z(prec, rnd);
    mpfr_pow (z.mpc_x, x.mpc_x, y.mpc_x, MPFR::frs2mprnd(rnd));
    return z;
    }


MPFR pow (const MPFR &x, const unsigned long int &y)
    {
    MPFR z(x.prec(), x.round_mode());
    mpfr_pow_ui (z.mpc_x, x.mpc_x, y, x.mpc_round_mode);
    return z;
    }


inline MPFR pow (const MPFR &x, const unsigned int &y) { return pow(x, static_cast<unsigned long int>(y)); }

MPFR pow (const unsigned long int &x, const MPFR &y)
    {
    MPFR z(y.prec(), y.round_mode());
    mpfr_ui_pow (z.mpc_x, x, y.mpc_x, y.mpc_round_mode);
    return z;
    }


inline MPFR pow (const unsigned int &x, const MPFR &y) { return pow(static_cast<unsigned long int>(x), y); }

MPFR pow (const MPFR &x, const long int &y)
    {
    MPFR z(x.prec(), x.round_mode());
    mpfr_pow_si (z.mpc_x, x.mpc_x, y, x.mpc_round_mode);
    return z;
    }


inline MPFR pow (const MPFR &x, const int &y) { return pow(x, static_cast<long int>(y)); }

MPFR pow (const MPFR &x, const mpz_t &y)
    {
    MPFR z(x.prec(), x.round_mode());
    mpfr_pow_z (z.mpc_x, x.mpc_x, y, x.mpc_round_mode);
    return z;
    }


MPFR pow (const unsigned long int &x, const unsigned long int &y, mp_prec_t prec=MPFR::default_prec(), std::float_round_style rnd=MPFR::default_round_mode())
    {
    MPFR z(prec, rnd);
    mpfr_ui_pow_ui (z.mpc_x, x, y, MPFR::frs2mprnd(rnd));
    return z;
    }


MPFR pow (const unsigned long int &x, const unsigned int &y, mp_prec_t prec=MPFR::default_prec(), std::float_round_style rnd=MPFR::default_round_mode())
    {
    MPFR z(prec, rnd);
    mpfr_ui_pow_ui (z.mpc_x, x, y, MPFR::frs2mprnd(rnd));
    return z;
    }


//------------------------------------------------------------
// Increment and decrement
//------------------------------------------------------------

MPFR MPFR::operator-- (int) { MPFR y = *this; mpfr_sub_ui (mpc_x, mpc_x, 1, mpc_round_mode); return y; }
MPFR MPFR::operator++ (int) { MPFR y = *this; mpfr_add_ui (mpc_x, mpc_x, 1, mpc_round_mode); return y; }
MPFR MPFR::operator-- () { mpfr_sub_ui (mpc_x, mpc_x, 1, mpc_round_mode); return *this; }
MPFR MPFR::operator++ () { mpfr_add_ui (mpc_x, mpc_x, 1, mpc_round_mode); return *this; }

template<typename T> inline void operator+= (MPFR &x, const T &y) { x = x + y; }
template<typename T> inline void operator-= (MPFR &x, const T &y) { x = x - y; }
template<typename T> inline void operator*= (MPFR &x, const T &y) { x = x * y; }
template<typename T> inline void operator/= (MPFR &x, const T &y) { x = x / y; }

#endif                                            // __BASIC_ARITHMETIC_CPP
